////////////////////////////////////////////////////////////////////////////////
///                             ir_machine.c                                 ///
///    IR remote control decoder with automatic protocol detection           ///
///                                                                          ///
/// Copyright 2017 Jaakko Kairus                                             ///
/// jaakko@kair.us                                                           ///
///                                                                          ///
/// You may freely use and modify this software for personal use.            ///
/// Commercial use of this software or parts of it is prohibited             ///
/// without written permission from author.                                  ///
/// DISCLAIMER: This software comes with ABSOLUTELY NO WARRANTY              ///
///                                                                          ///
/// For more information about this project see project web page at          ///
/// http://kair.us/projects/clock/v4/index.html                              ///
///                                                                          ///
/// Developed with CCS compiler version 5.076                                ///
///                                                                          ///
////////////////////////////////////////////////////////////////////////////////

#ifndef _PCH
#byte STATUS = 0x03
#else
#byte STATUS = 0x0FD8
#endif
#bit CARRY = STATUS.0 

#define NEC_BIT_TIME 18  // Bit time in 64us units = 1.152ms (should be 1.12ms)
#define NEC_PULSE_MAX 11 // Max pulse width in 64us units = 704us (nominal is 560 s)
#define NEC_REPEAT_INTERVAL 1891	// Interval between messages. Nominally 110 ms, 121 ms / 64 s = 1891
#define SIRC_BIT_THRESHOLD 14  // Separator between 0 (600us) and 1 (1200us). 14*64=896s
#define SIRC_REPEAT_INTERVAL 781	// Interval between messages. Nominally 45 ms, 50 ms / 64 s = 781
#define JVC_REPEAT_INTERVAL 1031	// Nominally 50...60 ms. 66 ms / 64 s = 1031
#define RC5_REPEAT_INTERVAL 1031	// Nominally 114 ms. 125 ms / 64 s = 1953


typedef enum {
        IR_IDLE,
        IR_START_PULSE,
        IR_START_PAUSE,
        IR_NEC_REPETITION,
        IR_NEC_DATA_PULSE,
        IR_NEC_DATA_PAUSE,
		IR_NEC_DATA_FINISHED,
		IR_NEC_DECODE,
		IR_NEC42_DECODE,
		IR_ERROR,
		IR_JVC_DECODE,
		IR_RC5_HALF_OF_BIT,
		IR_RC5_DATA_1,
		IR_RC5_DATA_0,
		IR_RC5_DECODE,
		IR_SIRC_DATA_PULSE,
		IR_SIRC_DATA_PAUSE,
		IR_SIRC_DATA_FINISHED,
		IR_SIRC_DECODE
} IR_STATES;
static IR_STATES IR_state = IR_IDLE;

static unsigned int16 actcycles = 0;		// How many timer cycles since last keypress, is the repetition still valid? 
static unsigned int8 repeats=0;
static unsigned int32 ir_receive_buffer;
static unsigned int16 ir_receive_buffer_16;
static unsigned int8 DIGIT1;
static unsigned int8 DIGIT2;
static unsigned int8 DIGIT3;
static unsigned int8 DIGIT4;
static unsigned int8 BIT_TIMER;
static unsigned int8 new_packet = 0;
static unsigned int8 is_samsung;
static unsigned int8 mark_bit;
static unsigned int8 bits_received;

typedef struct commonIrPacket {
	unsigned int8 protocol;		// protocol, e.g. NEC_PROTOCOL
	unsigned int16 address;		// address
	unsigned int16 command;		// command
	unsigned int8 flags;		// flags, e.g. repetition
} IR_PACKET;

void ir_machine(void);
int1 getIrPacket(IR_PACKET *completedPacket);

void ir_machine(void) {
	switch (IR_state) {
		case IR_IDLE:
			//if (new_packet)	// Do not receive new ir before previous 
			//	return FALSE;	// has been procssed
			if (actcycles!=0)
				actcycles--;
			else
				repeats=0;
			
			if (input(IRin)) // Input high (inactive)?
				return;
			BIT_TIMER=0;
			IR_state=IR_START_PULSE;
			mark_bit=0;
			bits_received=0;
			return;
		


		case IR_START_PULSE:
			BIT_TIMER++;
			if (!(input(IRin))) { // Pulse active?
				if (BIT_TIMER>155) { // Longer than longest supported protocol (9 ms + 10%) low?
					IR_state=IR_ERROR;
				}
			}
			else {
				if (BIT_TIMER>118 && BIT_TIMER<155) {  // 8.4ms -10% to 9ms + 10% low? (7560/64, 9900/64)
					IR_state=IR_START_PAUSE;		  // NEC or JVC
					is_samsung=0;
					BIT_TIMER=0;
				}
				else if (BIT_TIMER>63 && BIT_TIMER<78) {  // 4.5ms +/- 10% low?
					IR_state=IR_START_PAUSE;		  // Samsung
					is_samsung=1;
					BIT_TIMER=0;
				}
				else if (BIT_TIMER>33 && BIT_TIMER<41) { // 2.4m +/- 10% low?
					IR_state=IR_SIRC_DATA_PAUSE;
					BIT_TIMER=0;
				}	
				else if (BIT_TIMER>24 && BIT_TIMER<32) {	// RC-5 extended
					BIT_TIMER=0;
					shift_left(&ir_receive_buffer, 4, 0);
					bits_received++;
					IR_state=IR_RC5_HALF_OF_BIT;
				}	
				else if (BIT_TIMER>10 && BIT_TIMER<18) {
					IR_state=IR_RC5_HALF_OF_BIT;	// RC-5
				}	
				else if (BIT_TIMER>5 && BIT_TIMER<11) {
					IR_state = IR_NEC_DATA_PAUSE;	// JVC repeat. Go to NEC routine and determine first data bit
					is_samsung=0;
				}	
				else {
					IR_state=IR_IDLE;	  // too short, go back to idle
				}
			}
			return;
		case IR_START_PAUSE:			// Measure pause after first AGC pulse
			BIT_TIMER++;
			if (input(IRin)) {			// Input high?
				if (BIT_TIMER>79) {		// Longer than 4.5ms + 10% high?
					IR_state=IR_IDLE;		// Yes, not good! Start from state 0.
				}
			}
			else {						// Input is low.. 
				if (BIT_TIMER>52) {		// Let's see if the gap was long (>3.35ms)
					IR_state=IR_NEC_DATA_PULSE;			// It was long, go to state 4 and decode message
				}
				else {
					IR_state=IR_NEC_REPETITION;			// It was a short one! It's a repetition!
				}
				BIT_TIMER=0;
			}
			return;
		case IR_NEC_REPETITION:
			IR_state=IR_ERROR;
			if (actcycles!=0) {   // Is the repetition valid?
				actcycles=NEC_REPEAT_INTERVAL;
				if (repeats<5)
					repeats++;
				if (repeats==0 || repeats==5)
					new_packet = 2;
			}
			return;
		case IR_NEC_DATA_PULSE:
			BIT_TIMER++;
			if (!(input(IRin))) {  // Input still low?
				if (BIT_TIMER>NEC_PULSE_MAX) { // See if max pulse width reached
					IR_state=IR_ERROR; 			// Max width reached, goto error state to wait it gets high
				}
			}
			else {
				IR_state=IR_NEC_DATA_PAUSE;		// Input is now high
			}
			return;	
		case IR_NEC_DATA_PAUSE:
			BIT_TIMER++;
			if (BIT_TIMER < NEC_BIT_TIME*3) { // Should we keep waiting?
				if (input(IRin)) { // Yes. Is the input still high?
				}
				else {
					if (bits_received<32)
						shift_right(&ir_receive_buffer, 4, (BIT_TIMER>NEC_BIT_TIME*3/2+1));	// Use 1.5 bit time as threshold
					else
						shift_right(&ir_receive_buffer_16, 2, (BIT_TIMER>NEC_BIT_TIME*3/2+1));	// Use 1.5 bit time as threshold
					bits_received++;
					IR_state = IR_NEC_DATA_PULSE;
					BIT_TIMER = 0;
				}
			}
			else {  // Data has been ended.
				if (bits_received == 16)
					IR_state = IR_JVC_DECODE; // Check if we have valid JVC code
				else if (bits_received == 32)
					IR_state = IR_NEC_DECODE;
				else if (bits_received == 42)
					IR_state = IR_NEC42_DECODE;
				else
					IR_state = IR_ERROR;
			}
			return;
		case IR_NEC_DECODE:
			DIGIT4 = ir_receive_buffer;
			DIGIT3 = ir_receive_buffer>>8;
			DIGIT2 = ir_receive_buffer>>16;
			DIGIT1 = ir_receive_buffer>>24;
			
			IR_state=IR_IDLE;

			actcycles=NEC_REPEAT_INTERVAL;		

			repeats++;

			if (repeats==1 || repeats==5) {
				if (is_samsung)
					new_packet = 3;
				else
					new_packet = 2;
			}		
			if (repeats==5)
				repeats=4;
			return;
		case IR_NEC42_DECODE:
			DIGIT4 = ir_receive_buffer;				// address 8 lower bits
			DIGIT3 = ir_receive_buffer>>8 & 0x1f;	// address 5 upper bits
			DIGIT2 = (ir_receive_buffer>>26) + (ir_receive_buffer_16&0xc0);	// command
			DIGIT1 = ir_receive_buffer_16>>8;		// !command
			
			if ((DIGIT1 ^ DIGIT2) != 0xff) {	// verify command with its complement
				IR_state=IR_ERROR;
				return;
			}	
				
			IR_state=IR_IDLE;

			actcycles=NEC_REPEAT_INTERVAL;		

			repeats++;

			if (repeats==1 || repeats==5) {
				new_packet = 28;
			}		
			if (repeats==5)
				repeats=4;
			return;
		case IR_ERROR:
			if (input(IRin)) { // Return to idle state when receiver goes idle
				IR_state=IR_IDLE;
			}
			return;
		case IR_JVC_DECODE:
			DIGIT2 = ir_receive_buffer>>24;
			DIGIT3 = 0;
			DIGIT4 = ir_receive_buffer>>16;
			
			IR_state=IR_IDLE;
			actcycles=JVC_REPEAT_INTERVAL;
			repeats++;

			if (repeats==1 || repeats==8)
				new_packet = 20;	// JVC protocol
			if (repeats==9)
				repeats=7;
			return;

		case IR_RC5_HALF_OF_BIT:
			if (bits_received==13) {
				IR_state = IR_RC5_DECODE;
				return;
			}	
			BIT_TIMER++;				// Wait until center of first half of bit
			if (BIT_TIMER<20)
				return;	// Not yet!
			if (input(IRin))	// If input is high, go to state 3
				IR_state=IR_RC5_DATA_1;
			else				// If it's low, go to state 4
				IR_state=IR_RC5_DATA_0;
			return;
		case IR_RC5_DATA_1:	// was 7839 bytes with asm routines on this and following two states
			BIT_TIMER++;	// now 7901 with C routines, but also handles RC5 extended protcol
			if (BIT_TIMER>50) {
				IR_state = IR_ERROR;	// Took too long
				return;
			}	
			if (input(IRin))	// Is it still high?
				return;			// We will wait...
			
			shift_left(&ir_receive_buffer, 4, 1);
			bits_received++;
			BIT_TIMER = 0;
			IR_state = IR_RC5_HALF_OF_BIT;				
			return;
		case IR_RC5_DATA_0:
			BIT_TIMER++;
			if (BIT_TIMER>50) {
				IR_state = IR_ERROR;	// Took too long
				return;
			}	
			if (!(input(IRin)))	// Is it still low?
				return;			// We will wait...

			shift_left(&ir_receive_buffer, 4, 0);
			bits_received++;
			BIT_TIMER = 0;
			IR_state = IR_RC5_HALF_OF_BIT;
			return;			
		case IR_RC5_DECODE:
			DIGIT4=(ir_receive_buffer>>6)&0x1f;	// 5-bit address
			DIGIT2=ir_receive_buffer&0x3f;	// 6-bit command
			if (!(ir_receive_buffer&0x1000))	// 7th command bit
				DIGIT2+=0x40;
			DIGIT3=0;
			DIGIT1= 0;	
			IR_state = IR_ERROR;		
			actcycles=RC5_REPEAT_INTERVAL;		
			repeats++;
			if (repeats==1 || repeats==5) {
				new_packet = 7;
			}		
			if (repeats==5)
				repeats=4;
			return;
		case IR_SIRC_DATA_PULSE:
			BIT_TIMER++;
			if (BIT_TIMER < SIRC_BIT_THRESHOLD*2) { // Should we keep waiting?
				if (!(input(IRin))) { // Yes. Is the input still low?
					return;  // It is - we'll wait..
				}
				else {
					shift_right(&ir_receive_buffer, 4, (BIT_TIMER>SIRC_BIT_THRESHOLD));	// Use 1.5 bit time as threshold
					bits_received++;
					IR_state = IR_SIRC_DATA_PAUSE;
					BIT_TIMER = 0;
               		
					return;
				}
			}
			else {  // Pulse is too long. It is not SIRC
				IR_state = IR_ERROR;
				return;
			}
		case IR_SIRC_DATA_PAUSE:
			BIT_TIMER++;
			if (input(IRin)) {  // Input still high?
				if (BIT_TIMER>SIRC_BIT_THRESHOLD) { // See if max pause width reached
					if (bits_received == 12 || bits_received == 15 || bits_received == 20)	// It can also be actual message end...
						IR_state = IR_SIRC_DATA_FINISHED;
					else
						IR_state=IR_ERROR; 			// Max width reached, goto error state to wait it gets high
				}
			}
			else {
				IR_state=IR_SIRC_DATA_PULSE;		// Input is now low
				BIT_TIMER = 0;
			}
			return;	
		case IR_SIRC_DATA_FINISHED:	// Additional wait that nothing is following in 4*900us = 3.6 ms
			BIT_TIMER++;
			if (input(IRin)) {
				if (BIT_TIMER>SIRC_BIT_THRESHOLD*4) {
					IR_state = IR_SIRC_DECODE;
				}
			}
			else {
				IR_state=IR_ERROR;
			}
			return;
		case IR_SIRC_DECODE:
/*
			if (bits_received==12) {
				DIGIT3=0;
				DIGIT4=ir_receive_buffer>>27;	// 5-bit address
				DIGIT2=(ir_receive_buffer>>20)&0x7f;	// 7-bit data
			}	
			else if (bits_received==15) {
				DIGIT3=0;
				DIGIT4=ir_receive_buffer>>24;	// 8-bit address
				DIGIT2=(ir_receive_buffer>>17)&0x7f;	// 7-bit data
			}
			else {
				DIGIT3=ir_receive_buffer>>27;	// 5-bit extra
				DIGIT4=ir_receive_buffer>>19;	// 8-bit address
				DIGIT2=(ir_receive_buffer>>12)&0x7f;	// 7-bit data
			}	
			DIGIT1= 0;			
*/
			DIGIT3=ir_receive_buffer>>(47-bits_received);	// 5-bit extra
			DIGIT4=ir_receive_buffer>>(39-bits_received);	// 8-bit address
			DIGIT2=(ir_receive_buffer>>(32-bits_received))&0x7f;	// 7-bit data
			DIGIT1= 0;			

			IR_state=IR_IDLE;		// We're done! Let's start from beginning
			actcycles=SIRC_REPEAT_INTERVAL;

			repeats++;

			if (repeats==1 || repeats==10)
				new_packet = 1;
			if (repeats==11)
				repeats=9;

			return;

//		default:	// Should not get here but let's get out anyway
//			IR_state = IR_IDLE;
//			return;
	}
}





int1 getIrPacket(IR_PACKET *completedPacket) {
	if (new_packet) {
		completedPacket->protocol = new_packet;
		completedPacket->address = DIGIT3*256+DIGIT4;
		completedPacket->command = DIGIT2;
		completedPacket->flags = 0x00;
		new_packet=0;
		return TRUE;
	}
	else
		return FALSE;	
}	